import numpy as np

try:
    import cupy
except:
    pass
from ._topologies import generateInitialConfiguration
from .src.distutils import PartialDistance
from .src.core import (
    Encode2ElasticMatrix,
    MakeUniformElasticMatrix,
    PrimitiveElasticGraphEmbedment,
    PrimitiveElasticGraphEmbedment_cp,
)
from .src.BaseElPi import computeElasticPrincipalGraph


def computeElasticPrincipalGraphWithGrammars(
    X,
    GrowGrammars,
    ShrinkGrammars,
    NumNodes,
    NumEdges=float("inf"),
    InitNodes=3,
    Lambda=0.01,
    Mu=0.1,
    GrammarOptimization=False,
    MaxSteps=float("inf"),
    GrammarOrder=["Grow", "Shrink"],
    MaxNumberOfIterations=10,
    TrimmingRadius=float("inf"),
    eps=0.01,
    Do_PCA=True,
    InitNodePositions=None,
    AdjustVect=None,
    ElasticMatrix=None,
    InitEdges=None,
    CenterData=True,
    ComputeMSEP=True,
    verbose=False,
    ShowTimer=False,
    ReduceDimension=None,
    #                                             drawAccuracyComplexity = True,
    #                                             drawPCAView = True,
    #                                             drawEnergy = True,
    n_cores=1,
    # ClusType = "Sock",
    MinParOp=20,
    nReps=1,
    #                                              ParallelRep = False,
    Subsets=list(),
    ProbPoint=1,
    PointWeights=None,
    Mode=1,
    FinalEnergy="Base",
    alpha=0,
    beta=0,
    # gamma = 0,
    # FastSolve = False,
    Configuration="Line",
    DensityRadius=None,
    AvoidSolitary=False,
    EmbPointProb=1,
    SampleIC=True,
    AvoidResampling=True,
    AdjustElasticMatrix=None,
    AdjustElasticMatrix_Initial=None,
    Lambda_Initial=None,
    Mu_Initial=None,
    DisplayWarnings=True,
    StoreGraphEvolution=False,
    GPU=False,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    Construct a principal graph with the specified grammar

    This function is a wrapper to the computeElasticPrincipalGraph function that constructs the appropriate initial graph and
    apply the required grammar operations. Note that this is a generic function that is called by the topology specific functions.

    X numerical 2D matrix, the n-by-m matrix with the position of n m-dimensional points
    NumNodes integer, the number of nodes of the principal graph
    Lambda real, the lambda parameter used the compute the elastic energy
    Mu real, the lambda parameter used the compute the elastic energy
    InitNodes integer, number of points to include in the initial graph
    MaxNumberOfIterations integer, maximum number of steps to embed the nodes in the data
    TrimmingRadius real, maximal distance of point from a node to affect its embedment
    eps real, minimal relative change in the position of the nodes to stop embedment
    Do_PCA boolean, should data and initial node positions be PCA trnasformed?
    InitNodePositions numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
    in the initial step
    InitEdges numerical 2D matrix, the e-by-2 matrix with e end-points of the edges connecting the nodes
    ElasticMatrix numerical 2D matrix, the k-by-k elastic matrix
    CenterData boolean, should data and initial node positions be centered?
    ComputeMSEP boolean, should MSEP be computed when building the report?
    verbose boolean, should debugging information be reported?
    ShowTimer boolean, should the time to construct the graph be computed and reported for each step?
    ReduceDimension integer vector, vector of principal components to retain when performing
    dimensionality reduction. If NULL all the components will be used
    drawAccuracyComplexity boolean, should the accuracy VS complexity plot be reported?
    drawPCAView boolean, should a 2D plot of the points and pricipal curve be dranw for the final configuration?
    drawEnergy boolean, should changes of evergy VS the number of nodes be reported?
    n.cores either an integer (indicating the number of cores to used for the creation of a cluster) or
    cluster structure returned, e.g., by makeCluster. If a cluster structure is used, all the nodes must contains X
    (this is done using clusterExport)
    MinParOP integer, the minimum number of operations to use parallel computation
    nReps integer, number of replica of the construction
    ProbPoint real between 0 and 1, probability of inclusing of a single point for each computation
    Subsets list of column names (or column number). When specified a principal tree will be computed for each of the subsets specified.
    NumEdges integer, the maximum nulber of edges
    Mode integer, the energy computation mode
    FastSolve boolean, should FastSolve be used when fitting the points to the data?
    ClusType string, the type of cluster to use. It can gbe either "Sock" or "Fork".
    Currently fork clustering only works in Linux
    Configuration string, initial configuration type.
    DensityRadius numeric, the radius used to estimate local density. This need to be set when ICOver is equal to "Density"
    AvoidSolitary boolean, should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    FinalEnergy string indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha positive numeric, the value of the alpha parameter of the penalized elastic energy
    beta positive numeric, the value of the beta parameter of the penalized elastic energy
    EmbPointProb numeric between 0 and 1. If less than 1 point will be sampled at each iteration.
    EmbPointProb indicates the probability of using each points. This is an *experimental* feature, which may
    helps speeding up the computation if a large number of points is present.
    GrowGrammars list of strings, the grammar to be used in the growth step
    ShrinkGrammars list of strings, the grammar to be used in the shrink step
    SampleIC boolean, should the initial configuration be considered on the sampled points when applicable?
    AdjustVect boolean vector keeping track of the nodes for which the elasticity parameters have been adjusted.
    When true for a node its elasticity parameters will not be adjusted.
    gamma
    AdjustElasticMatrix a penalization function to adjust the elastic matrices after a configuration has been chosen (e.g., AdjustByConstant).
    If NULL (the default), no penalization will be used.
    AdjustElasticMatrix.Initial a penalization function to adjust the elastic matrices of the initial configuration (e.g., AdjustByConstant).
    If NULL (the default), no penalization will be used.
    Lambda.Initial real, the lambda parameter used the construct the elastic matrix associted with ther initial configuration if needed.
    If NULL, the value of Lambda will be used.
    Mu.Initial real, the mu parameter used the construct the elastic matrix associted with ther initial configuration if needed.
    If NULL, the value of Mu will be used.
    GrammarOptimization boolean, should grammar optimization be perfomred? If true grammar operations that do not increase the number of
    nodes will be allowed
    MaxSteps integer, max number of applications of the grammar. This value need to be less than infinity if GrammarOptimization is set to true
    GrammarOrder character vector, the order of application of the grammars. It can be any combination of "Grow" and "Shrink"
    AvoidResampling booleand, should the sampling of initial conditions avoid reselecting the same points
    (or points neighbors if DensityRadius is specified)?

    Return
    -------
    A list of principal graph strucutures containing the trees constructed during the different replica of the algorithm.
    If the number of replicas is larger than 1. The the final element of the list is the "average tree", which is constructed by
    fitting the coordinates of the nodes of the reconstructed trees


    @examples


    """
    # Be default we are using a predefined initial configuration
    ComputeIC = False

    if InitNodes<3:
        raise ValueError('InitNodes must be >=3')
    # Generate a dummy subset is not specified
    if Subsets == list():
        Subsets = [np.array(range(X.shape[1]))]

    # Prepare the list to be returned
    ReturnList = list()

    # Copy the original matrix, this is needed in case of subsetting (and setting float64 dtype to avoid numba compilation errors)
    Base_X = X.astype("float64")

    # For each subset
    for j in range(len(Subsets)):

        # Generate the appropriate matrix
        X = Base_X[:, Subsets[j]]
        SquaredX = np.sum(X ** 2, axis=1, keepdims=1)
        if GPU:
            Xcp = cupy.asarray(X)
            SquaredXcp = Xcp.sum(axis=1, keepdims=1)

        # Define temporary variable to avoid excessing plotting
        # Intermediate_drawPCAView = drawPCAView
        # Intermediate_drawAccuracyComplexity = drawAccuracyComplexity
        # Intermediate_drawEnergy = drawEnergy

        Used = np.array([False] * len(X))

        for i in range(nReps):

            # Select the points to be used
            if ProbPoint < 1 and ProbPoint > 0:
                SelPoints = np.random.random(len(X)) <= ProbPoint
            else:
                SelPoints = np.array([True] * len(X))

            # Do we need to compute the initial conditions?
            if InitNodePositions is None or (
                InitEdges is None and ElasticMatrix is None
            ):
                if verbose:
                    print("Generating the initial configuration")

                # We are computing the initial conditions. InitNodePositions need to be reset after each step!
                ComputeIC = True

                if SampleIC:
                    if AvoidResampling:
                        InitialConf = generateInitialConfiguration(
                            X[SelPoints & ~Used, :],
                            Nodes=InitNodes,
                            Configuration=Configuration,
                            DensityRadius=DensityRadius,
                            verbose=verbose,
                        )

                        Dist = np.min(
                            PartialDistance(InitialConf["NodePositions"], X),
                            axis=0,
                        )

                        if DensityRadius:
                            Used = Used | (Dist < DensityRadius)
                        else:
                            Used = Used | (Dist <= np.finfo(float).min)

                        if (np.sum(Used) < len(X) * 0.9) and verbose:
                            print(
                                "90% of the points have been used as initial"
                                " conditions. Resetting."
                            )
                    else:
                        # Construct the initial configuration
                        InitialConf = generateInitialConfiguration(
                            X[SelPoints, :],
                            Nodes=InitNodes,
                            Configuration=Configuration,
                            DensityRadius=DensityRadius,
                            verbose=verbose,
                        )

                else:
                    if AvoidResampling:
                        InitialConf = generateInitialConfiguration(
                            X[
                                ~Used,
                            ],
                            Nodes=InitNodes,
                            Configuration=Configuration,
                            DensityRadius=DensityRadius,
                            verbose=verbose,
                        )

                        Dist = np.min(
                            PartialDistance(InitialConf["NodePositions"], X),
                            axis=0,
                        )

                        if DensityRadius:
                            Used = Used | (Dist < DensityRadius)
                        else:
                            Used = Used | (Dist < np.finfo(float).min)

                        if (np.sum(Used) > len(X) * 0.9) and verbose:
                            print(
                                "90% or more of the points have been used as"
                                " initial conditions. Resetting."
                            )

                    else:
                        # Construct the initial configuration
                        InitialConf = generateInitialConfiguration(
                            X,
                            Nodes=InitNodes,
                            Configuration=Configuration,
                            DensityRadius=DensityRadius,
                            verbose=verbose,
                        )

                # Set the initial edge configuration
                InitEdges = InitialConf["Edges"]

                # Compute the initial elastic matrix
                InitialElasticMatrix = MakeUniformElasticMatrix(
                    Edges=InitialConf["Edges"], Lambda=Lambda, Mu=Mu
                )
                # Compute the initial node position
                if GPU:
                    InitNodePositions = PrimitiveElasticGraphEmbedment_cp(
                        X=X,
                        NodePositions=InitialConf["NodePositions"],
                        MaxNumberOfIterations=MaxNumberOfIterations,
                        TrimmingRadius=TrimmingRadius,
                        eps=eps,
                        ElasticMatrix=InitialElasticMatrix,
                        Mode=Mode,
                        Xcp=Xcp,
                        SquaredXcp=SquaredXcp,
                        SquaredX=SquaredX,
                        FixNodesAtPoints=[],
                        PointWeights=PointWeights,
                    )[0]
                else:
                    InitNodePositions = PrimitiveElasticGraphEmbedment(
                        X=X,
                        NodePositions=InitialConf["NodePositions"],
                        MaxNumberOfIterations=MaxNumberOfIterations,
                        TrimmingRadius=TrimmingRadius,
                        eps=eps,
                        ElasticMatrix=InitialElasticMatrix,
                        Mode=Mode,
                        SquaredX=SquaredX,
                        FixNodesAtPoints=[],
                        PointWeights=PointWeights,
                    )[0]

            # Do we need to compute AdjustVect?
            if AdjustVect is None:
                if FixNodesAtPoints != []:
                    AdjustVect = [False] * (
                        len(InitNodePositions) + len(FixNodesAtPoints)
                    )
                else:
                    AdjustVect = [False] * len(InitNodePositions)

            # Limit plotting after a few examples
            # if(len(ReturnList) == 3):
            #    print("Graphical output will be suppressed for the remaining replicas")
            #    Intermediate_drawPCAView = False
            #    Intermediate_drawAccuracyComplexity = False
            #    Intermediate_drawEnergy = False

            if verbose:
                print(
                    "Constructing tree",
                    i + 1,
                    "of",
                    nReps,
                    "/ Subset",
                    j + 1,
                    "of",
                    len(Subsets),
                )
            # Run the ElPiGraph algorithm
            ReturnList.append(
                computeElasticPrincipalGraph(
                    Data=X[SelPoints, :],
                    NumNodes=NumNodes,
                    NumEdges=NumEdges,
                    InitNodePositions=InitNodePositions,
                    InitEdges=InitEdges,
                    ElasticMatrix=ElasticMatrix,
                    AdjustVect=AdjustVect,
                    GrowGrammars=GrowGrammars,
                    ShrinkGrammars=ShrinkGrammars,
                    GrammarOptimization=GrammarOptimization,
                    MaxSteps=MaxSteps,
                    GrammarOrder=GrammarOrder,
                    MaxNumberOfIterations=MaxNumberOfIterations,
                    TrimmingRadius=TrimmingRadius,
                    eps=eps,
                    Lambda=Lambda,
                    Mu=Mu,
                    Do_PCA=Do_PCA,
                    CenterData=CenterData,
                    ComputeMSEP=ComputeMSEP,
                    verbose=verbose,
                    ShowTimer=ShowTimer,
                    ReduceDimension=ReduceDimension,
                    Mode=Mode,
                    FinalEnergy=FinalEnergy,
                    alpha=alpha,
                    beta=beta,  # gamma = gamma,
                    # drawAccuracyComplexity = Intermediate_drawAccuracyComplexity,
                    # drawPCAView = Intermediate_drawPCAView,
                    # drawEnergy = Intermediate_drawEnergy,
                    n_cores=n_cores,
                    # ClusType = ClusType,
                    MinParOp=MinParOp,
                    # FastSolve = FastSolve,
                    AvoidSolitary=AvoidSolitary,
                    EmbPointProb=EmbPointProb,
                    PointWeights=PointWeights,
                    AdjustElasticMatrix=AdjustElasticMatrix,
                    AdjustElasticMatrix_Initial=AdjustElasticMatrix_Initial,
                    Lambda_Initial=Lambda_Initial,
                    Mu_Initial=Mu_Initial,
                    DisplayWarnings=DisplayWarnings,
                    StoreGraphEvolution=StoreGraphEvolution,
                    GPU=GPU,
                    FixNodesAtPoints=FixNodesAtPoints,
                    pseudotime=pseudotime,
                    pseudotimeLambda=pseudotimeLambda,
                    label=label,
                    labelLambda=labelLambda,
                    MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
                )
            )

            # Save extra information
            ReturnList[-1]["SubSetID"] = j
            ReturnList[-1]["ReplicaID"] = i
            ReturnList[-1]["ProbPoint"] = ProbPoint

            # Reset InitNodePositions for the next iteration
            if ComputeIC:
                InitNodePositions = None

    # Are we using bootstrapping (nReps > 1). If yes we compute the consensus tree
    if nReps > 1:
        if verbose:
            print("Constructing average tree")

        # The nodes of the principal trees will be used as points to compute the consensus tree
        AllPoints = np.concatenate(([i["NodePositions"] for i in ReturnList]))

        # De we need to compute the initial conditions?
        if InitNodePositions is None or (
            InitEdges is None and ElasticMatrix is None
        ):

            # construct the initial configuration
            InitialConf = generateInitialConfiguration(
                AllPoints,
                Nodes=InitNodes,
                Configuration=Configuration,
                DensityRadius=DensityRadius,
                verbose=verbose,
            )

            # print(InitialConf)

            # Set the initial edge configuration
            InitEdges = InitialConf["Edges"]
            # Compute the initial elastic matrix
            EM = Encode2ElasticMatrix(Edges=InitEdges, Lambdas=Lambda, Mus=Mu)

            # Compute the initial node position
            if GPU:
                InitNodePositions = PrimitiveElasticGraphEmbedment_cp(
                    X=X,
                    NodePositions=InitialConf["NodePositions"],
                    MaxNumberOfIterations=MaxNumberOfIterations,
                    TrimmingRadius=TrimmingRadius,
                    eps=eps,
                    ElasticMatrix=EM,
                    Mode=Mode,
                    Xcp=Xcp,
                    SquaredXcp=SquaredXcp,
                    PointWeights=PointWeights,
                )[0]

            else:
                InitNodePositions = PrimitiveElasticGraphEmbedment(
                    X=X,
                    NodePositions=InitialConf["NodePositions"],
                    MaxNumberOfIterations=MaxNumberOfIterations,
                    TrimmingRadius=TrimmingRadius,
                    eps=eps,
                    ElasticMatrix=EM,
                    Mode=Mode,
                    PointWeights=PointWeights,
                )[0]

        ReturnList.append(
            computeElasticPrincipalGraph(
                Data=AllPoints,
                NumNodes=NumNodes,
                NumEdges=NumEdges,
                InitNodePositions=InitNodePositions,
                InitEdges=InitEdges,
                ElasticMatrix=ElasticMatrix,
                AdjustVect=AdjustVect,
                GrowGrammars=GrowGrammars,
                ShrinkGrammars=ShrinkGrammars,
                MaxNumberOfIterations=MaxNumberOfIterations,
                TrimmingRadius=TrimmingRadius,
                eps=eps,
                Lambda=Lambda,
                Mu=Mu,
                Do_PCA=Do_PCA,
                CenterData=CenterData,
                ComputeMSEP=ComputeMSEP,
                verbose=verbose,
                ShowTimer=ShowTimer,
                ReduceDimension=None,
                Mode=Mode,
                FinalEnergy=FinalEnergy,
                alpha=alpha,
                beta=beta,  # gamma = gamma,
                # drawAccuracyComplexity = drawAccuracyComplexity,
                # drawPCAView = drawPCAView, drawEnergy = drawEnergy,
                n_cores=n_cores,
                # ClusType = ClusType,
                MinParOp=MinParOp,
                # FastSolve = FastSolve,
                AvoidSolitary=AvoidSolitary,
                EmbPointProb=EmbPointProb,
                PointWeights=PointWeights,
                AdjustElasticMatrix=AdjustElasticMatrix,
                AdjustElasticMatrix_Initial=AdjustElasticMatrix_Initial,
                Lambda_Initial=Lambda_Initial,
                Mu_Initial=Mu_Initial,
                DisplayWarnings=DisplayWarnings,
                StoreGraphEvolution=StoreGraphEvolution,
                GPU=GPU,
                FixNodesAtPoints=FixNodesAtPoints,
                pseudotime=pseudotime,
                pseudotimeLambda=pseudotimeLambda,
                label=label,
                labelLambda=labelLambda,
                MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
            )
        )

        # Run the ElPiGraph algorithm
        ReturnList[-1]["SubSetID"] = j
        ReturnList[-1]["ReplicaID"] = 0
        ReturnList[-1]["ProbPoint"] = 1

    return ReturnList



import numpy as np
import numba as nb
import igraph
from .graphs import GetSubGraph

# def ComputePrimitiveGraphElasticEnergy(NodePositions, ElasticMatrix, dists):
#     MSE = dists.sum() / np.size(dists)
#     Mu = ElasticMatrix.diagonal()
#     Lambda = np.triu(ElasticMatrix, 1)
#     StarCenterIndices = np.nonzero(Mu > 0)[0]
#     (row, col) = Lambda.nonzero()
#     dev = NodePositions.take(row, axis=0) - NodePositions.take(col, axis=0)
#     L = Lambda[Lambda > 0]
#     EP = sum(L.flatten('F')*np.sum(dev**2, axis=1))
#     indL = Lambda+Lambda.transpose() > 0
#     RP = 0
#     for i in range(np.size(StarCenterIndices)):
#         leaves = indL.take(StarCenterIndices[i], axis=1)
#         K = sum(leaves)
#         dev = (NodePositions.take(StarCenterIndices[i], axis=0) -
#                sum(NodePositions[leaves])/K)
#         RP = RP + Mu[StarCenterIndices[i]] * sum(dev ** 2)
#     ElasticEnergy = MSE + EP + RP
#     return ElasticEnergy, MSE, EP, RP


@nb.njit(cache=True)
def ComputePrimitiveGraphElasticEnergy(
    NodePositions, ElasticMatrix, dists, PointWeights=None
):
    """
        //' Compute the elastic energy associated with a particular configuration
    //'
    //' This function computes the elastic energy associated to a set of points and graph embedded
    //' into them.
    //'
    //' NodePositions A numeric k-by-m matrix containing the position of the k nodes of the embedded graph
    //' ElasticMatrix A numeric l-by-l matrix containing the elastic parameters associates with the edge
    //' of the embedded graph
    //' dists A numeric vector containind the squared distance of the data points to the closest node of the graph
    //'
    //' Return
    -------
    A list with four elements:
    //' * ElasticEnergy is the total energy
    //' * MSE is the MSE component of the energy
    //' * EP is the EP component of the energy
    //' * RP is the RP component of the energy
    """

    MSE = np.sum(dists * PointWeights) / np.sum(PointWeights)

    Mu = np.diag(ElasticMatrix)
    Lambda = np.triu(ElasticMatrix, 1)
    StarCenterIndices = (Mu > 0).nonzero()[0]
    (row, col) = Lambda.nonzero()
    dev = NodePositions[row] - NodePositions[col]

    dev2 = np.sum(dev ** 2, axis=1)
    L = np.zeros((len(row)))
    for i in range(len(row)):
        L[i] = Lambda[row[i], col[i]]
    EP = np.dot(L, dev2)

    indL = (Lambda + Lambda.transpose()) > 0
    RP = 0
    for i in range(StarCenterIndices.size):
        leaves = indL[StarCenterIndices[i]]
        ind_leaves = leaves.nonzero()[0]
        K = ind_leaves.size
        dev_ = NodePositions[StarCenterIndices[i]] - (
            NodePositions[ind_leaves] / K
        ).sum(axis=0)
        RP = RP + Mu[StarCenterIndices[i]] * (dev_ ** 2).sum()
    ElasticEnergy = MSE + EP + RP
    return ElasticEnergy, MSE, EP, RP


def RadialCount(A, B, A_squared, Dvect):
    Dvect = Dvect ** 2
    distances = (
        -2 * np.matmul(A, np.transpose(B))
        + np.sum(np.square(B), axis=1)
        + A_squared
    )

    idx = np.array([np.where(distances < dist)[0] for dist in Dvect])
    count = np.array([len(i) for i in idx])
    return count, idx[-1]


# def ComputePenalizedPrimitiveGraphElasticEnergy(NodePositions, ElasticMatrix, dists,alpha=.1,beta=.1):
#     MSE = dists.sum() / np.size(dists)
#     Mu = ElasticMatrix.diagonal()
#     Lambda = np.triu(ElasticMatrix, 1)
#     StarCenterIndices = np.nonzero(Mu > 0)[0]
#     (row, col) = Lambda.nonzero()
#     dev = NodePositions.take(row, axis=0) - NodePositions.take(col, axis=0)
#     L = Lambda[Lambda > 0]
#     ### diff compared to base function
#     BinEM = ElasticMatrix.copy()
#     np.fill_diagonal(BinEM,0)
#     BinEM[BinEM > 0] = 1
#     Ks = np.sum(BinEM,axis=0)
#     lp = np.maximum(Ks[row], Ks[col])
#     lp = lp-2
#     lp[lp<0] = 0
#     Lpenalized = L + alpha*lp
#     EP = sum(Lpenalized * np.sum(dev**2, axis=1))
#     # ###
#     indL = Lambda+Lambda.transpose() > 0
#     RP = 0
#     for i in range(np.size(StarCenterIndices)):
#         leaves = indL.take(StarCenterIndices[i], axis=1)
#         K = sum(leaves)
#         dev = (NodePositions.take(StarCenterIndices[i], axis=0) -
#                sum(NodePositions[leaves])/K)
#         RP += Mu[StarCenterIndices[i]]* np.power(K,beta) * sum(dev ** 2)

#     ElasticEnergy = MSE + EP + RP

#     return ElasticEnergy, MSE, EP, RP


@nb.njit(cache=True)
def ComputePenalizedPrimitiveGraphElasticEnergy(
    NodePositions, ElasticMatrix, dists, alpha=0.01, beta=0.01
):
    """
        //' Compute the penalized elastic energy associated with a particular configuration
    //'
    //' This function computes the elastic energy associated to a set of points and graph embedded
    //' into them.
    //'
    //' NodePositions A numeric k-by-m matrix containing the position of the k nodes of the embedded graph
    //' ElasticMatrix A numeric l-by-l matrix containing the elastic parameters associates with the edge
    //' of the embedded graph
    //' dists A numeric vector containing the squared distance of the data points to the closest node of the graph
    //' alpha
    //' beta
    //'
    //' Return
    -------
    A list with four elements:
    //' * ElasticEnergy is the total energy
    //' * MSE is the MSE component of the energy
    //' * EP is the EP component of the energy
    //' * RP is the RP component of the energy
    """
    MSE = dists.sum() / dists.size
    Mu = np.diag(ElasticMatrix)
    Lambda = np.triu(ElasticMatrix, 1)
    StarCenterIndices = (Mu > 0).nonzero()[0]
    (row, col) = Lambda.nonzero()
    dev = NodePositions[row] - NodePositions[col]
    L = np.zeros((len(row)))
    for i in range(len(row)):
        L[i] = Lambda[row[i], col[i]]
    ### diff compared to base function
    BinEM = (Lambda + Lambda.transpose()) > 0
    Ks = BinEM.sum(axis=0)
    lp = np.maximum(Ks[row], Ks[col])
    lp = lp - 2
    lp[np.where(lp < 0)] = 0

    Lpenalized = L + alpha * lp
    EP = np.dot(Lpenalized, np.sum(dev ** 2, axis=1))
    ####
    indL = Lambda + Lambda.transpose() > 0
    RP = 0
    for i in range(StarCenterIndices.size):
        leaves = indL[StarCenterIndices[i]]
        ind_leaves = leaves.nonzero()[0]
        K = ind_leaves.size
        dev_ = NodePositions[StarCenterIndices[i]] - (
            NodePositions[ind_leaves] / K
        ).sum(axis=0)
        RP += Mu[StarCenterIndices[i]] * (K ** beta) * (dev_ ** 2).sum()

    ElasticEnergy = MSE + EP + RP

    return ElasticEnergy, MSE, EP, RP


@nb.njit(cache=True)
def ComputePenalizedPrimitiveGraphElasticEnergy_v2(
    NodePositions,
    ElasticMatrix,
    dists,
    alpha=0.01,
    beta=0.01,
    PointWeights=None,
    PseudotimeNodePositions=None,
):
    """
        //' Compute the penalized elastic energy associated with a particular configuration
    //'
    //' This function computes the elastic energy associated to a set of points and graph embedded
    //' into them.
    //'
    //' NodePositions A numeric k-by-m matrix containing the position of the k nodes of the embedded graph
    //' ElasticMatrix A numeric l-by-l matrix containing the elastic parameters associates with the edge
    //' of the embedded graph
    //' dists A numeric vector containing the squared distance of the data points to the closest node of the graph
    //' alpha
    //' beta
    //'
    //' Return
    -------
    A list with four elements:
    //' * ElasticEnergy is the total energy
    //' * MSE is the MSE component of the energy
    //' * EP is the EP component of the energy
    //' * RP is the RP component of the energy
    """
    MSE = np.sum(dists * PointWeights) / np.sum(PointWeights)

    Mu = np.diag(ElasticMatrix)
    Lambda = np.triu(ElasticMatrix, 1)
    StarCenterIndices = (Mu > 0).nonzero()[0]
    (row, col) = Lambda.nonzero()
    dev = NodePositions[row] - NodePositions[col]
    L = np.zeros((len(row)))
    for i in range(len(row)):
        L[i] = Lambda[row[i], col[i]]
    ### diff compared to base function
    BinEM = (Lambda + Lambda.transpose()) > 0
    Ks = BinEM.sum(axis=0)
    lp = np.maximum(Ks[row], Ks[col])
    lp = lp - 2
    lp[np.where(lp < 0)] = 0

    Lpenalized = L + alpha * lp
    EP = np.dot(Lpenalized, np.sum(dev ** 2, axis=1))

    if PseudotimeNodePositions is not None:
        ps_dev = NodePositions - PseudotimeNodePositions
        pseudotimeEP = np.dot(
            np.repeat(L.min(), len(NodePositions)), np.sum(ps_dev ** 2, axis=1)
        )
        EP += pseudotimeEP

    ####
    indL = Lambda + Lambda.transpose() > 0
    RP = 0
    for i in range(StarCenterIndices.size):
        leaves = indL[StarCenterIndices[i]]
        ind_leaves = leaves.nonzero()[0]
        K = ind_leaves.size
        dev_ = NodePositions[StarCenterIndices[i]] - (
            NodePositions[ind_leaves] / K
        ).sum(axis=0)
        RP += Mu[StarCenterIndices[i]] * (K ** beta) * (dev_ ** 2).sum()

    ElasticEnergy = MSE + EP + RP

    return ElasticEnergy, MSE, EP, RP


def ComputePenalizedPrimitiveGraphElasticEnergy_v3(
    NodePositions,
    ElasticMatrix,
    part,
    dists,
    alpha=0.01,
    beta=0.01,
    PseudotimeNodePositions=None,
    Label=None,
    LabelLambda=None,
):
    """
        //' Compute the penalized elastic energy associated with a particular configuration
    //'
    //' This function computes the elastic energy associated to a set of points and graph embedded
    //' into them.
    //'
    //' NodePositions A numeric k-by-m matrix containing the position of the k nodes of the embedded graph
    //' ElasticMatrix A numeric l-by-l matrix containing the elastic parameters associates with the edge
    //' of the embedded graph
    //' dists A numeric vector containing the squared distance of the data points to the closest node of the graph
    //' alpha
    //' beta
    //'
    //' Return
    -------
    A list with four elements:
    //' * ElasticEnergy is the total energy
    //' * MSE is the MSE component of the energy
    //' * EP is the EP component of the energy
    //' * RP is the RP component of the energy
    """
    MSE = dists.sum() / dists.size
    Mu = np.diag(ElasticMatrix)
    Lambda = np.triu(ElasticMatrix, 1)
    StarCenterIndices = (Mu > 0).nonzero()[0]
    (row, col) = Lambda.nonzero()
    dev = NodePositions[row] - NodePositions[col]
    L = np.zeros((len(row)))
    for i in range(len(row)):
        L[i] = Lambda[row[i], col[i]]
    ### diff compared to base function
    BinEM = (Lambda + Lambda.transpose()) > 0
    Ks = BinEM.sum(axis=0)
    lp = np.maximum(Ks[row], Ks[col])
    lp = lp - 2
    lp[np.where(lp < 0)] = 0

    Lpenalized = L + alpha * lp
    EP = np.dot(Lpenalized, np.sum(dev ** 2, axis=1))

    # ---pseudotime penalty
    if PseudotimeNodePositions is not None:
        ps_dev = NodePositions - PseudotimeNodePositions
        pseudotimeEP = np.dot(
            np.repeat(L.min(), len(NodePositions)), np.sum(ps_dev ** 2, axis=1)
        )
        EP += pseudotimeEP

    ####
    indL = Lambda + Lambda.transpose() > 0
    RP = 0
    for i in range(StarCenterIndices.size):
        leaves = indL[StarCenterIndices[i]]
        ind_leaves = leaves.nonzero()[0]
        K = ind_leaves.size
        dev_ = NodePositions[StarCenterIndices[i]] - (
            NodePositions[ind_leaves] / K
        ).sum(axis=0)
        RP += Mu[StarCenterIndices[i]] * (K ** beta) * (dev_ ** 2).sum()

    ElasticEnergy = MSE + EP + RP

    # ---categorical penalty
    if Label is not None:
        try:
            LL = LabelLoss(Label, LabelLambda, part, ElasticMatrix)
        except Exception as e:
            print(e)
            import pdb

            pdb.set_trace()
        ElasticEnergy = ElasticEnergy * (1 + LL)

    return ElasticEnergy, MSE, EP, RP


@nb.njit(cache=True)
def sum_squares_2d_array_along_axis1(arr):
    res = np.empty(arr.shape[0], dtype=arr.dtype)
    for o_idx in range(arr.shape[0]):
        sum_ = 0
        for i_idx in range(arr.shape[1]):
            sum_ += arr[o_idx, i_idx] * arr[o_idx, i_idx]
        res[o_idx] = sum_
    return res


@nb.njit(cache=True)
def euclidean_distance_square_numba(x1, x2):
    distances = np.sqrt(
        -2 * np.dot(x1, x2.T)
        + np.expand_dims(sum_squares_2d_array_along_axis1(x1), axis=1)
        + sum_squares_2d_array_along_axis1(x2)
    )
    return distances


# def PartialDistance(c1,c2):
#     with np.errstate(invalid='ignore'):

#         distances = np.sqrt(-2 * np.matmul(c1, np.transpose(c2)) + np.sum(np.square(c2), axis=1)
#                         + np.expand_dims(np.sum(np.square(c1), axis=1), axis=1))
#         distances[np.isnan(distances)]=0
#     return distances


def PartialDistance(A, B):
    a = euclidean_distance_square_numba(A, B)
    a[np.isnan(a)] = 0
    return a


def ComputeWeightedAverage(X, partition, PointWeights, NumberOfNodes):
    X = X * PointWeights
    # Auxiliary calculations
    M = X.shape[1]
    part = partition.ravel() + 1
    # Calculate total weights
    TotalWeight = PointWeights.sum()
    # Calculate weights for Relative size
    tmp = np.bincount(
        part, weights=PointWeights.ravel(), minlength=NumberOfNodes + 1
    )
    NodeClusterRelativeSize = tmp[1:] / TotalWeight
    # To prevent dividing by 0
    tmp[tmp == 0] = 1
    NodeClusterCenters = np.zeros((NumberOfNodes + 1, X.shape[1]))
    for k in range(M):
        NodeClusterCenters[:, k] = (
            np.bincount(part, weights=X[:, k], minlength=NumberOfNodes + 1)
            / tmp
        )
    return (
        NodeClusterCenters[
            1:,
        ],
        NodeClusterRelativeSize[np.newaxis].T,
    )


def FitGraph2DataGivenPartition(
    X, PointWeights, SpringLaplacianMatrix, partition
):
    """
    # Solves the SLAU to find new node positions
    """
    NumberOfNodes = SpringLaplacianMatrix.shape[0]
    NodeClusterCenters, NodeClusterRelativeSize = ComputeWeightedAverage(
        X, partition, PointWeights, NumberOfNodes
    )
    SLAUMatrix = (
        np.diag(NodeClusterRelativeSize.transpose()[0]) + SpringLaplacianMatrix
    )
    NewNodePositions = np.linalg.solve(
        SLAUMatrix, NodeClusterRelativeSize * NodeClusterCenters
    )
    return NewNodePositions


def FitSubGraph2DataGivenPartition(
    move_X,
    move_PointWeights,
    SpringLaplacianMatrix,
    NodePositions,
    move_partition,
    move_nodes_idx,
):
    """
    Fits moving subpart of the graph to data while constraining some nodes to remain fixed
    """

    # params
    NumberOfNodes = len(NodePositions)
    NumberOfNodesToFit = len(move_nodes_idx)
    fixed_nodes_idx = list(set(range(NumberOfNodes)) - set(move_nodes_idx))
    # new node positions
    NewNodePositions = np.zeros(NodePositions.shape)

    # weighted average of moving nodes
    (
        move_NodeClusterCenters,
        move_NodeClusterRelativeSize,
    ) = ComputeWeightedAverage(
        move_X, move_partition, move_PointWeights, NumberOfNodesToFit
    )

    # SLAUMatrices
    rs = np.zeros((NumberOfNodes))
    rs[move_nodes_idx] = move_NodeClusterRelativeSize.flatten()
    SLAUMatrix = np.diag(rs) + SpringLaplacianMatrix
    move_SLAUMatrix = SLAUMatrix[move_nodes_idx][:, move_nodes_idx]
    fixed_SLAUMatrix = SLAUMatrix[move_nodes_idx][:, fixed_nodes_idx]

    # Fit
    RightHandSide = move_NodeClusterRelativeSize * move_NodeClusterCenters
    rhs1 = NodePositions[fixed_nodes_idx]
    RightHandSide -= np.dot(fixed_SLAUMatrix, rhs1)
    move_NewNodePositions = np.linalg.solve(move_SLAUMatrix, RightHandSide)

    # Store
    NewNodePositions[fixed_nodes_idx, :] = NodePositions[fixed_nodes_idx, :]
    NewNodePositions[move_nodes_idx, :] = move_NewNodePositions

    return NewNodePositions


def FitSubGraph2DataGivenPartition_v2(
    move_X,
    move_PointWeights,
    SpringLaplacianMatrix,
    NodePositions,
    move_partition,
    move_nodes_idx,
    PseudotimeNodePositions,
    PseudotimeLambda,
):
    """
    Fits moving subpart of the graph to data while constraining some nodes to remain fixed
    """

    # params
    NumberOfNodes = len(NodePositions)
    NumberOfNodesToFit = len(move_nodes_idx)
    fixed_nodes_idx = list(set(range(NumberOfNodes)) - set(move_nodes_idx))
    # new node positions
    NewNodePositions = np.zeros(NodePositions.shape)

    # weighted average of moving nodes
    (
        move_NodeClusterCenters,
        move_NodeClusterRelativeSize,
    ) = ComputeWeightedAverage(
        move_X, move_partition, move_PointWeights, NumberOfNodesToFit
    )

    if PseudotimeNodePositions is not None:
        move_NodeClusterCenters = (
            move_NodeClusterCenters
            + PseudotimeNodePositions[move_nodes_idx] * PseudotimeLambda
        ) / (1 + PseudotimeLambda)

    # SLAUMatrices
    rs = np.zeros((NumberOfNodes))
    rs[move_nodes_idx] = move_NodeClusterRelativeSize.flatten()
    SLAUMatrix = np.diag(rs) + SpringLaplacianMatrix
    move_SLAUMatrix = SLAUMatrix[move_nodes_idx][:, move_nodes_idx]
    fixed_SLAUMatrix = SLAUMatrix[move_nodes_idx][:, fixed_nodes_idx]

    # Fit
    RightHandSide = move_NodeClusterRelativeSize * move_NodeClusterCenters
    rhs1 = NodePositions[fixed_nodes_idx]
    RightHandSide -= np.dot(fixed_SLAUMatrix, rhs1)
    move_NewNodePositions = np.linalg.solve(move_SLAUMatrix, RightHandSide)

    # Store
    NewNodePositions[fixed_nodes_idx, :] = NodePositions[fixed_nodes_idx, :]
    NewNodePositions[move_nodes_idx, :] = move_NewNodePositions

    return NewNodePositions


#####################################
### Categorical supervision utils ###
#####################################


# @nb.njit
def normalized_entropy_times_cardinality(num_labels):
    """Calculates the normalized entropy of labels

    Parameters
    ----------
    num_labels : int array, shape = [n_samples]
        The labels
    """
    if len(num_labels) == 0:
        return 1.0
    pi = np.bincount(num_labels).astype(np.float64)
    pi = pi[pi > 0]
    pi_sum = np.sum(pi)
    # log(a / b) should be calculated as log(a) - log(b) for
    # possible loss of precision
    return (
        -np.sum((pi / pi_sum) * (np.log(pi) - np.log(pi_sum)))
        / np.log(len(num_labels))
        * len(num_labels)
    )


# @nb.njit
def LabelLoss(Label, LabelLambda, part, EM):
    valid_nodes = np.unique(part)
    n_nodes = int(np.sum(valid_nodes >= 0))
    _EM = EM.copy()
    np.fill_diagonal(_EM, 0)
    net = igraph.Graph.Adjacency((_EM > 0).tolist(), mode="undirected")
    branches = GetSubGraph(net, "branches")
    branches_dataidx = [np.isin(part, b) for b in branches]

    entropy = 0
    for i in range(len(branches)):
        entropy += normalized_entropy_times_cardinality(
            Label[branches_dataidx[i]]
        )
        # normalize 0/1
    LL = LabelLambda * entropy / len(part)
    # ElasticEn * (1 + LabelLoss)
    return LL


"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
from distutils import util
from os import path
from io import open


here = path.abspath(path.dirname(__file__))
path_src = util.convert_path("elpigraph/src")
# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Get version
ver_path = util.convert_path("elpigraph/_version.py")
main_ns = {}
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)
VERSION = main_ns["__version__"]
# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name="elpigraph-python",  # Required
    version=VERSION,  # Required
    description="",  # Optional
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",  # Optional (see note above)
    url="https://github.com/j-bac/elpigraph-python",  # Optional
    maintainer="Jonathan Bac",  # Optional
    maintainer_email="",  # Optional
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Pick your license as you wish
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    keywords="machine_learning graphs dimension_reduction single_cell",  # Optional
    packages=["elpigraph.src", "elpigraph"],
    package_dir={
        "elpigraph": "elpigraph",
        "elpigraph.src": path_src,
    },
    #     package_data={'': ['data/']},
    install_requires=[
        "numpy >=1.16.2",
        "pandas >=0.23.4",
        "numba >=0.49.1 ",
        "scikit-learn >=0.21.3",
        "scipy >=1.2.0",
        "python_igraph >=0.7.1",
        "networkx >=2.0",
        "matplotlib",
        "shapely",
    ],
    project_urls={  # Optional
        "Bug Reports": "https://github.com/j-bac/elpigraph-python/issues",
        "Source": "https://github.com/j-bac/elpigraph-python/",
    },
    zip_safe=False,
    extras_require={
        "tests": ["pytest", "pytest-cov"],
        "docs": [
            "sphinx",
            "sphinx-gallery",
            "sphinx_rtd_theme",
            "numpydoc",
            "matplotlib",
        ],
    },
)



# -*- coding: utf-8 -*-
"""
Created on Fri Feb  9 10:11:02 2018

@author: Alexis Martin
"""
import numpy as np

try:
    import cupy
except:
    pass
from scipy import linalg as la
from sklearn.decomposition import TruncatedSVD
from sklearn.utils import check_random_state


def PCA(data):
    """
    returns: data transformed in 2 dims/columns + regenerated original data
    pass in: data as 2D NumPy array
    """
    # mean center the data
    # data -= data.mean(axis=0)
    # calculate the covariance matrix
    R = np.cov(data, rowvar=False)
    # calculate eigenvectors & eigenvalues of the covariance matrix
    # use 'eigh' rather than 'eig' since R is symmetric,
    # the performance gain is substantial
    evals, evecs = np.linalg.eigh(R)
    # sort eigenvalue in decreasing order
    idx = np.argsort(evals)[::-1]
    evecs = evecs[:, idx]
    # sort eigenvectors according to same index
    evals = evals[idx]
    # carry out the transformation on the data using eigenvectors
    # and return the re-scaled data, eigenvalues, and eigenvectors

    return evecs, data.dot(evecs), evals


def TruncPCA(X, n_components, algorithm="arpack"):
    svd = TruncatedSVD(algorithm=algorithm, n_components=n_components)
    prcomp = svd.fit_transform(X)
    s = svd.singular_values_
    Vt = svd.components_
    U = prcomp / s

    return prcomp, svd.explained_variance_, U, s, Vt


def PCA_gpu(data):
    """
    returns: data transformed in 2 dims/columns + regenerated original data
    pass in: data as 2D NumPy array
    """
    # mean center the data
    # data -= data.mean(axis=0)
    # calculate the covariance matrix
    R = cupy.cov(data, rowvar=False)
    # calculate eigenvectors & eigenvalues of the covariance matrix
    # use 'eigh' rather than 'eig' since R is symmetric,
    # the performance gain is substantial
    evals, evecs = cupy.linalg.eigh(R)
    # sort eigenvalue in decreasing order
    idx = cupy.argsort(evals)[::-1]
    evecs = evecs[:, idx]
    # sort eigenvectors according to same index
    evals = evals[idx]
    # carry out the transformation on the data using eigenvectors
    # and return the re-scaled data, eigenvalues, and eigenvectors

    return evecs, data.dot(evecs), evals


def TruncSVD_gpu(
    M, n_components, n_oversamples=10, n_iter="auto", transpose="auto", random_state=0
):
    """Computes a truncated randomized SVD on GPU. Adapted from Sklearn.
    Taken from : https://vip.readthedocs.io/en/latest/_modules/vip_hci/pca/svd.html

    Parameters
    ----------
    M : ndarray or sparse matrix
        Matrix to decompose
    n_components : int
        Number of singular values and vectors to extract.
    n_oversamples : int (default is 10)
        Additional number of random vectors to sample the range of M so as
        to ensure proper conditioning. The total number of random vectors
        used to find the range of M is n_components + n_oversamples. Smaller
        number can improve speed but can negatively impact the quality of
        approximation of singular vectors and singular values.
    n_iter : int or 'auto' (default is 'auto')
        Number of power iterations. It can be used to deal with very noisy
        problems. When 'auto', it is set to 4, unless `n_components` is small
        (< .1 * min(X.shape)) `n_iter` in which case is set to 7.
        This improves precision with few components.
    transpose : True, False or 'auto' (default)
        Whether the algorithm should be applied to M.T instead of M. The
        result should approximately be the same. The 'auto' mode will
        trigger the transposition if M.shape[1] > M.shape[0] since this
        implementation of randomized SVD tend to be a little faster in that
        case.
    random_state : int, RandomState instance or None, optional (default=None)
        The seed of the pseudo random number generator to use when shuffling
        the data.  If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random number
        generator; If None, the random number generator is the RandomState
        instance used by `np.random`.
    lib : {'cp', 'pytorch'}, str optional
        Chooses the GPU library to be used.

    Notes
    -----
    This algorithm finds a (usually very good) approximate truncated
    singular value decomposition using randomization to speed up the
    computations. It is particularly fast on large matrices on which
    you wish to extract only a small number of components. In order to
    obtain further speed up, `n_iter` can be set <=2 (at the cost of
    loss of precision).

    References
    ----------
    * Finding structure with randomness: Stochastic algorithms for constructing
      approximate matrix decompositions
      Halko, et al., 2009 http://arxiv.org/abs/arXiv:0909.4061
    * A randomized algorithm for the decomposition of matrices
      Per-Gunnar Martinsson, Vladimir Rokhlin and Mark Tygert
    * An implementation of a randomized algorithm for principal component
      analysis
      A. Szlam et al. 2014
    """
    random_state = check_random_state(random_state)
    n_random = n_components + n_oversamples
    n_samples, n_features = M.shape

    if n_iter == "auto":
        # Checks if the number of iterations is explicitly specified
        n_iter = 7 if n_components < 0.1 * min(M.shape) else 4

    if transpose == "auto":
        transpose = n_samples < n_features
    if transpose:
        M = M.T  # this implementation is a bit faster with smaller shape[1]

    # Generating normal random vectors with shape: (M.shape[1], n_random)
    Q = random_state.normal(size=(M.shape[1], n_random))
    Q = cupy.array(Q)
    Q = cupy.asarray(Q)

    # Perform power iterations with Q to further 'imprint' the top
    # singular vectors of M in Q
    for i in range(n_iter):
        Q = cupy.dot(M, Q)
        Q = cupy.dot(M.T, Q)

    # Sample the range of M using by linear projection of Q. Extract an orthonormal basis
    Q, _ = cupy.linalg.qr(cupy.dot(M, Q), mode="reduced")

    # project M to the (k + p) dimensional space using the basis vectors
    B = cupy.dot(Q.T, M)

    B = cupy.array(B)
    Q = cupy.array(Q)
    # compute the SVD on the thin matrix: (k + p) wide
    Uhat, s, V = cupy.linalg.svd(B, full_matrices=False, compute_uv=True)
    del B
    U = cupy.dot(Q, Uhat)

    if transpose:
        # transpose back the results according to the input convention
        return V[:n_components, :].T, s[:n_components], U[:, :n_components].T
    else:
        return U[:, :n_components], s[:n_components], V[:n_components, :]



import numpy as np
from .core import DecodeElasticMatrix, PartitionData
from .distutils import ComputePrimitiveGraphElasticEnergy


def getPrimitiveGraphStructureBarCode(ElasticMatrix):
    # Mus = ElasticMatrix.diagonal()
    Lambda = ElasticMatrix.copy()
    np.fill_diagonal(Lambda, 0)
    indL = Lambda > 0
    Connectivities = indL.sum(axis=0)
    Mcon = np.max(Connectivities)

    counts = np.bincount(Connectivities)[1:]
    code = "||" + str(ElasticMatrix.shape[0])

    if Mcon <= 2:
        code = "0" + code
    else:
        code = "|".join([str(c) for c in counts[2:][::-1]]) + code
    return code


def project_point_onto_graph(X, NodePositions, Edges, Partition=None):
    """
    Project data points on the precipal graph

    X numerical matrix containg points on the rows and dimensions on the columns
    NodePositions numerical matrix containg the positions of the nodes on the rows
    (must have the same dimensionality of X)
    Edges a 2-dimensional matrix containing edges as pairs of integers. The integers much
    match the rows of NodePositions
    Partition a Partition vector associating points to at most one of the nodes of the graph.
    It can be NULL, in which case it will be computed by the algorithm

    Return
    -------
    A list with several elements:
    \itemize{
     \item{"X_projected "}{A matrix containing the projection of the points (on rows) on the edges of the graph}
     \item{"MSEP "}{The mean squared error (distance) of the points from the graph}
     \item{"ProjectionValues "}{The normalized position of the point on its associted edge.
     A value <0 indicates a projection before the initial position of the node.
     A value >1 indicates a projection after the final position of the node.
     A value betwen 0 and 1 indicates at which percentage of the edge length the point is being projected,
     e.g., a value of 0.3 indicates the 30\%.}
     \item{"EdgeID "}{An integer indicating the id of the edge on which each point has been projected. Note that
     if a point is projected on a node, this id will indicate one of the edges connected to that node.}
     \item{"EdgeLen "}{The length of the edges described by the Edges input matrix}
     \item{"NodePositions "}{the NodePositions input matrix}
     \item{"Edges "}{the Edges input matrix}
    }



    @examples
    """
    if Partition is None:
        Partition = PartitionData(
            X,
            NodePositions,
            MaxBlockSize=10000,
            SquaredX=np.sum(X ** 2, axis=1, keepdims=1),
        )[0]

    X_projected = np.zeros(X.shape)
    ProjectionValues = np.array([np.inf] * len(X))
    Distances_squared = np.array([np.inf] * len(X))
    EdgeID = np.array([np.nan] * len(X))
    EdgeLen = np.array([np.inf] * len(Edges))

    for i in range(len(Edges)):
        Idxs = np.where(np.isin(Partition, Edges[i, :]))[0]

        PrjStruct = project_point_onto_edge(
            X=X[Idxs, :],
            NodePositions=NodePositions[
                Edges[
                    i,
                ],
                :,
            ],
            Edge=np.array([0, 1]),
        )

        if len(Idxs) > 0:
            ToFill = PrjStruct["Distance_Squared"] < Distances_squared[Idxs]
            X_projected[Idxs[ToFill],] = PrjStruct["X_Projected"][
                ToFill,
            ]
            try:
                ProjectionValues[Idxs[ToFill]] = PrjStruct["Projection_Value"][
                    ToFill
                ]
            except:
                ### only one index and proj value
                ProjectionValues[Idxs[ToFill]] = np.array(
                    [PrjStruct["Projection_Value"]]
                )[ToFill]

            Distances_squared[Idxs[ToFill]] = PrjStruct["Distance_Squared"][
                ToFill
            ]
            EdgeID[Idxs[ToFill]] = i

        EdgeLen[i] = np.sqrt(PrjStruct["EdgeLen_Squared"])

    return dict(
        X_projected=X_projected,
        MSEP=np.mean(Distances_squared),
        ProjectionValues=ProjectionValues,
        EdgeID=EdgeID,
        EdgeLen=EdgeLen,
        NodePositions=NodePositions,
        Edges=Edges,
    )


def project_point_onto_edge(X, NodePositions, Edge, ExtProj=False):
    """
    Title

    X
    NodePositions
    Edge

    @return


    @examples
    """

    vec = (NodePositions[Edge[1], :] - NodePositions[Edge[0], :])[:, None]
    u = ((X.T - NodePositions[Edge[0]][:, None]).T @ vec) / (vec.T @ vec)
    u[~np.isfinite(u)] = 0
    u = u.squeeze()
    X_Projected = X.copy()
    X_Projected[:] = np.nan

    if np.any(u < 0):
        X_Projected[u < 0, :] = np.repeat(
            NodePositions[Edge[0], :][None], np.sum(u < 0), axis=0
        )

    if np.any(u > 1):
        X_Projected[u > 1, :] = np.repeat(
            NodePositions[Edge[1], :][None], np.sum(u > 1), axis=0
        )

    if ExtProj:
        OnEdge = np.array([True] * len(u))
    else:
        OnEdge = (u >= 0) & (u <= 1)

    if np.any(OnEdge):
        UExp = np.reshape(
            np.repeat(u[OnEdge], len(vec)),
            (len(vec), int(np.sum(OnEdge))),
            order="F",
        )
        X_Projected[OnEdge, :] = (
            UExp * vec + NodePositions[Edge[0]][:, None]
        ).T

    distance_squared = np.sum((X_Projected - X) * (X_Projected - X), axis=1)

    return dict(
        X_Projected=X_Projected,
        Projection_Value=u,
        Distance_Squared=distance_squared,
        EdgeLen_Squared=np.sum(vec ** 2),
    )


def ReportOnPrimitiveGraphEmbedment(
    X,
    NodePositions,
    ElasticMatrix,
    PartData=None,
    ComputeMSEP=False,
    PointWeights=None,
):
    """
    # %   This function computes various measurements concerning a primitive
    # %   graph embedment
    # %
    # %           BARCODE is code in form ...S4|S3||N, where N is number of
    # %               nodes, S3 is number of 3-stars, S4 (S5,...) is number of
    # %               four (five,...) stars, etc.
    # %           ENERGY is total elastic energy of graph embedment (ENERGY = MSE + UE +
    #                                                                  %               UR)
    # %           NNODES is number of nodes.
    # %           NEDGES is number of edges
    # %           NRIBS is number of two stars (nodes with two otherr connected
    #                                           %               nodes).
    # %           NSTARS is number of stars with 3 and more leaves (nodes
    #                                                               %               connected with central node).
    # %           NRAYS2 is sum of rays minus doubled number of nodes.
    # %           MSE is mean square error or assessment of data approximation
    # %               quality.
    # %           MSEP is mean square error after piece-wise linear projection on the edges
    # %           FVE is fraction of explained variance. This value always
    # %               between 0 and 1. Greater value means higher quality of
    # %               data approximation.
    # %           FVEP is same as FVE but computed after piece-wise linear projection on the edges
    # %           UE is total sum of squared edge lengths.
    # %           UR is total sum of star deviations from harmonicity.
    # %           URN is UR * nodes
    # %           URN2 is UR * nodes^2
    # %           URSD is standard deviation of UR
    """

    if PointWeights is None:
        PointWeights = np.ones((len(X), 1))

    Mus = ElasticMatrix.diagonal()
    Lambda = ElasticMatrix.copy()
    np.fill_diagonal(Lambda, 0)
    indL = Lambda > 0
    Connectivities = indL.sum(axis=0)
    Mcon = np.max(Connectivities)
    counts = np.bincount(Connectivities)[1:]
    DecodedMat = DecodeElasticMatrix(ElasticMatrix)

    TotalVariance = np.sum(np.var(X, axis=0, ddof=1))
    BARCODE = getPrimitiveGraphStructureBarCode(ElasticMatrix)

    if PartData is None:
        PartData = PartitionData(
            X=X,
            NodePositions=NodePositions,
            MaxBlockSize=1000000,
            SquaredX=np.sum(X ** 2, axis=1, keepdims=1),
        )

    Energies = ComputePrimitiveGraphElasticEnergy(
        NodePositions=NodePositions,
        ElasticMatrix=ElasticMatrix,
        dists=PartData[1],
        PointWeights=PointWeights,
    )

    NNODES = len(NodePositions)
    NEDGES = len(DecodedMat[0])

    if len(counts) > 1:
        NRIBS = counts[1]
    else:
        NRIBS = 0

    if len(counts) > 2:
        NSTARS = counts[2]
    else:
        NSTARS = 0

    NRAYS = 0
    NRAYS2 = 0

    if ComputeMSEP:
        NodeProj = project_point_onto_graph(
            X,
            NodePositions=NodePositions,
            Edges=DecodedMat[0],
            Partition=PartData[0],
        )
        MSEP = NodeProj["MSEP"]
        FVEP = (TotalVariance - MSEP) / TotalVariance
    else:
        MSEP = np.nan
        FVEP = np.nan

    FVE = (TotalVariance - Energies[1]) / TotalVariance
    URN = Energies[-1] * NNODES
    URN2 = Energies[-1] * NNODES * NNODES
    URSD = 0

    return dict(
        BARCODE=BARCODE,
        ENERGY=Energies[0],
        NNODES=NNODES,
        NEDGES=NEDGES,
        NRIBS=NRIBS,
        NSTARS=NSTARS,
        NRAYS=NRAYS,
        NRAYS2=NRAYS2,
        MSE=Energies[1],
        MSEP=MSEP,
        FVE=FVE,
        FVEP=FVEP,
        UE=Energies[2],
        UR=Energies[-1],
        URN=URN,
        URN2=URN2,
        URSD=URSD,
    )
